# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a segment."""

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity
import pyvista as pv

from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.checks import check_ndarray_is_all_nan
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS
from ansys.geometry.core.shapes.curves.line import Line
from ansys.geometry.core.sketch.edge import SketchEdge


class SketchSegment(SketchEdge, Line):
    """
    Provides segment representation of a line.

    Parameters
    ----------
    start : Point2D
        Starting point of the line segment.
    end : Point2D
        Ending point of the line segment.
    plane : Plane, optional
        Plane containing the sketched circle, which is the global XY plane
        by default.
    """

    @check_input_types
    def __init__(
        self,
        start: Point2D,
        end: Point2D,
        plane: Plane = Plane(),
    ):
        """Initialize the ``SketchSegment`` class."""
        # Call SketchEdge init method
        SketchEdge.__init__(self)

        # Perform sanity checks on point values given
        check_ndarray_is_all_nan(start, "start")
        check_ndarray_is_all_nan(end, "end")

        # Assign values to start and end
        self._start = start
        self._end = end

        # Check segment points values and units
        if self._start == self._end:
            raise ValueError(
                "Parameters 'start' and 'end' have the same values. No segment can be created."
            )

        if not self._start.unit == self._end.unit:
            self._start.unit = self._end.unit = DEFAULT_UNITS.LENGTH

        # Call the Line init method
        self._init_primitive_line_from_plane(plane)

    def _init_primitive_line_from_plane(self, plane: Plane) -> None:
        """
        Initialize correctly the underlying primitive ``Line`` class.

        Parameters
        ----------
        plane : Plane
            Plane containing the sketched line.
        """
        # Find the global start and end points
        start_glb = plane.origin + Point3D(
            self.start[0] * plane.direction_x + self.start[1] * plane.direction_y,
            unit=self.start.base_unit,
        )
        end_glb = plane.origin + Point3D(
            self.end[0] * plane.direction_x + self.end[1] * plane.direction_y,
            unit=self.end.base_unit,
        )

        # Define the line direction
        line_direction = UnitVector3D.from_points(start_glb, end_glb)

        # Call the Line init method
        Line.__init__(self, start_glb, line_direction)

    @property
    def start(self) -> Point2D:
        """Starting point of the segment."""
        return self._start

    @property
    def end(self) -> Point2D:
        """Ending point of the segment."""
        return self._end

    @property
    def length(self) -> Quantity:
        """Length of the segment."""
        return np.sqrt(
            np.square(self._end.x - self._start.x) + np.square(self._end.y - self._start.y)
        )

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """
        VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        import numpy as np

        return pv.Line(
            np.array(
                [
                    self.start.x.m_as(DEFAULT_UNITS.LENGTH),
                    self.start.y.m_as(DEFAULT_UNITS.LENGTH),
                    0,
                ],
                dtype=np.float_,
            ),
            np.array(
                [self.end.x.m_as(DEFAULT_UNITS.LENGTH), self.end.y.m_as(DEFAULT_UNITS.LENGTH), 0],
                dtype=np.float_,
            ),
        )

    @check_input_types
    def __eq__(self, other: "SketchSegment") -> bool:
        """Equals operator for the ``SketchSegment`` class."""
        return self.start == other.start and self.end == other.end

    def __ne__(self, other: "SketchSegment") -> bool:
        """Not equals operator for the ``SketchSegment`` class."""
        return not self == other

    def plane_change(self, plane: "Plane") -> None:
        """
        Redefine the plane containing ``SketchSegment`` objects.

        Notes
        -----
        This implies that their 3D definition might suffer changes.

        Parameters
        ----------
        plane : Plane
            Desired new plane that is to contain the sketched segment.
        """
        # Reinitialize the line definition for the given plane
        self._init_primitive_line_from_plane(plane)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a torus."""

from functools import cached_property

from beartype import beartype as check_input_types
from beartype.typing import Tuple, Union
import numpy as np
from pint import Quantity

from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Z
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.measurements import Distance
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
    ParamUV,
)
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.typing import Real, RealSequence


class Torus(Surface):
    """
    Provides 3D torus representation.

    Parameters
    ----------
    origin : Union[~numpy.ndarray, RealSequence, Point3D],
        Centered origin of the torus.
    direction_x : Union[~numpy.ndarray, RealSequence, UnitVector3D, Vector3D]
        X-axis direction.
    direction_y : Union[~numpy.ndarray, RealSequence, UnitVector3D, Vector3D]
        Y-axis direction.
    major_radius : Union[Quantity, Distance, Real]
        Major radius of the torus.
    minor_radius : Union[Quantity, Distance, Real]
        Minor radius of the torus.
    """

    @check_input_types
    def __init__(
        self,
        origin: Union[np.ndarray, RealSequence, Point3D],
        major_radius: Union[Quantity, Distance, Real],
        minor_radius: Union[Quantity, Distance, Real],
        reference: Union[np.ndarray, RealSequence, UnitVector3D, Vector3D] = UNITVECTOR3D_X,
        axis: Union[np.ndarray, RealSequence, UnitVector3D, Vector3D] = UNITVECTOR3D_Z,
    ):
        """Initialize the ``Torus`` class."""
        self._origin = Point3D(origin) if not isinstance(origin, Point3D) else origin
        self._reference = (
            UnitVector3D(reference) if not isinstance(reference, UnitVector3D) else reference
        )
        self._axis = UnitVector3D(axis) if not isinstance(axis, UnitVector3D) else axis
        if not self._reference.is_perpendicular_to(self._axis):
            raise ValueError("Torus reference (dir_x) and axis (dir_z) must be perpendicular.")

        # Store values in base unit
        self._major_radius = (
            major_radius if isinstance(major_radius, Distance) else Distance(major_radius)
        )
        self._minor_radius = (
            minor_radius if isinstance(minor_radius, Distance) else Distance(minor_radius)
        )

    @property
    def origin(self) -> Point3D:
        """Origin of the torus."""
        return self._origin

    @property
    def major_radius(self) -> Quantity:
        """Semi-major radius of the torus."""
        return self._major_radius.value

    @property
    def minor_radius(self) -> Quantity:
        """Semi-minor radius of the torus."""
        return self._minor_radius.value

    @property
    def dir_x(self) -> UnitVector3D:
        """X-direction of the torus."""
        return self._reference

    @property
    def dir_y(self) -> UnitVector3D:
        """Y-direction of the torus."""
        return self.dir_z.cross(self.dir_x)

    @property
    def dir_z(self) -> UnitVector3D:
        """Z-direction of the torus."""
        return self._axis

    @property
    def volume(self) -> Quantity:
        """Volume of the torus."""
        return 2 * np.pi**2 * self._major_radius.value * self._minor_radius.value**2

    @property
    def surface_area(self) -> Quantity:
        """Surface_area of the torus."""
        return 4 * np.pi**2 * self._major_radius.value * self._minor_radius.value

    @check_input_types
    def __eq__(self, other: "Torus") -> bool:
        """Equals operator for the ``Torus`` class."""
        return (
            self._origin == other._origin
            and self._major_radius == other._major_radius
            and self._minor_radius == other._minor_radius
            and self._reference == other._reference
            and self._axis == other._axis
        )

    def transformed_copy(self, matrix: Matrix44) -> "Torus":
        """
        Create a transformed copy of the torus based on a transformation matrix.

        Parameters
        ----------
        matrix : Matrix44
            4x4 transformation matrix to apply to the torus.

        Returns
        -------
        Torus
            New torus that is the transformed copy of the original torus.
        """
        new_point = self.origin.transform(matrix)
        new_reference = self._reference.transform(matrix)
        new_axis = self._axis.transform(matrix)
        return Torus(
            new_point,
            self.major_radius,
            self.minor_radius,
            UnitVector3D(new_reference[0:3]),
            UnitVector3D(new_axis[0:3]),
        )

    def mirrored_copy(self) -> "Torus":
        """
        Create a mirrored copy of the torus along the y-axis.

        Returns
        -------
        Torus
            New torus that is a mirrored copy of the original torus.
        """
        return Torus(
            self.origin, self.major_radius, self.minor_radius, -self._reference, -self._axis
        )

    def evaluate(self, parameter: ParamUV) -> "TorusEvaluation":
        """
        Evaluate the torus at the given parameters.

        Parameters
        ----------
        parameter : ParamUV
            Parameters (u,v) to evaluate the torus at.

        Returns
        -------
        TorusEvaluation
            Resulting evaluation.
        """
        return TorusEvaluation(self, parameter)

    def parameterization(self) -> Tuple[Parameterization, Parameterization]:
        """
        Parameterize the torus surface as a tuple (U and V respectively).

        The U parameter specifies the longitude angle, increasing clockwise (east) about
        the axis (right-hand corkscrew law). It has a zero parameter at
        ``Geometry.Frame.DirX`` and a period of ``2*pi``.

        The V parameter specifies the latitude, increasing north, with a zero parameter
        at the equator. For the donut, where the major radius is greater
        than the minor radius, the range is [-pi, pi] and the
        parameterization is periodic. For a degenerate torus, the range is restricted
        accordingly and the parameterization is non-periodic.

        Returns
        -------
        Tuple[Parameterization, Parameterization]
            Information about how a torus's u and v parameters are parameterized, respectively.
        """
        u = Parameterization(ParamForm.PERIODIC, ParamType.CIRCULAR, Interval(0, 2 * np.pi))
        v = Parameterization(
            ParamForm.PERIODIC, ParamType.CIRCULAR, Interval(-np.pi / 2, np.pi / 2)
        )

        return (u, v)

    def project_point(self, point: Point3D) -> "TorusEvaluation":
        """
        Project a point onto the torus and evaluate the torus.

        Parameters
        ----------
        point : Point3D
            Point to project onto the torus.

        Returns
        -------
        TorusEvaluation
            Resulting evaluation.
        """
        vector1 = UnitVector3D.from_points(self.origin, point)
        u = np.arctan2(vector1.dot(self.dir_y), vector1.dot(self.dir_x))
        localX = np.cos(u) * self.dir_x + np.sin(u) * self.dir_y
        delta = self.major_radius.m * localX
        vector2 = vector1 - delta
        if self.major_radius.m >= self.minor_radius.m:
            v = np.arctan2(vector2.dot(self.dir_z), vector2.dot(localX))
            return TorusEvaluation(self, ParamUV(u, v))
        vector3 = vector1 + delta
        v1 = np.arctan2(vector2.dot(self.dir_z), vector2.dot(localX)), -np.pi
        v2 = np.arctan2(vector3.dot(self.dir_z), vector3.dot(localX)), -np.pi
        if np.power(
            np.linalg.norm((TorusEvaluation(self, ParamUV(u, v1)).position() - point)), 2
        ) < np.power(
            np.linalg.norm((TorusEvaluation(self, ParamUV(u + np.pi, v2)).position() - point)), 2
        ):
            return TorusEvaluation(self, ParamUV(u, v1))
        else:
            return TorusEvaluation(self, ParamUV(u + np.pi, v2))

    def contains_param(self, param_uv: ParamUV) -> bool:  # noqa: D102
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:  # noqa: D102
        raise NotImplementedError("contains_point() is not implemented.")


class TorusEvaluation(SurfaceEvaluation):
    """
    Evaluate the torus`` at given parameters.

    Parameters
    ----------
    Torus: ~ansys.geometry.core.shapes.surfaces.torus.Torus
        Torust to evaluate.
    parameter: ParamUV
        Parameters (u, v) to evaluate the torus at.
    """

    def __init__(self, torus: Torus, parameter: ParamUV) -> None:
        """Initiate the ``TorusEvaluation`` class."""
        self._torus = torus
        self._parameter = parameter
        self.cache = {}

    @property
    def torus(self) -> Torus:
        """Torus being evaluated."""
        return self._torus

    @property
    def parameter(self) -> ParamUV:
        """Parameter that the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """
        Position of the evaluation.

        Returns
        -------
        Point3D
            Point that lies on the torus at this evaluation.
        """
        return (
            self._torus.origin
            + (self._torus.major_radius.m + np.cos(self.parameter.v) * self._torus.minor_radius.m)
            * self.__cylinder_normal
            + np.sin(self.parameter.v) * self._torus.minor_radius.m * self._torus.dir_z
        )

    @cached_property
    def normal(self) -> UnitVector3D:
        """
        Normal to the surface.

        Returns
        -------
        UnitVector3D
            Normal unit vector to the torus at this evaluation.
        """
        return UnitVector3D(
            np.cos(self.parameter.v) * self.__cylinder_normal
            + np.sin(self.parameter.v) * self._torus.dir_z
        )

    @cached_property
    def __cylinder_normal(self) -> Vector3D:
        """
        Normal to the surface.

        Returns
        -------
        UnitVector3D
            Normal unit vector to the torus at this evaluation.
        """
        return (
            np.cos(self.parameter.u) * self._torus.dir_x
            + np.sin(self.parameter.u) * self._torus.dir_y
        )

    @cached_property
    def __cylinder_tangent(self) -> Vector3D:
        """Private tangent helper method."""
        return (
            -np.sin(self.parameter.u) * self._torus.dir_x
            + np.cos(self.parameter.u) * self._torus.dir_y
        )

    @cached_property
    def u_derivative(self) -> Vector3D:
        """
        First derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the U parameter.
        """
        return (
            self._torus.major_radius.m + np.cos(self.parameter.v) * self._torus.minor_radius.m
        ) * self.__cylinder_tangent

    @cached_property
    def v_derivative(self) -> Vector3D:
        """
        First derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the V parameter.
        """
        return (
            -np.sin(self.parameter.v) * self._torus.minor_radius.m * self.__cylinder_tangent
            + np.cos(self.parameter.v) * self._torus.minor_radius.m * self._torus.dir_z
        )

    @cached_property
    def uu_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            Second derivative with respect to the U parameter.
        """
        return (
            -(self._torus.major_radius.m + np.cos(self.parameter.v))
            * self._torus.minor_radius.m
            * self.__cylinder_normal
        )

    @cached_property
    def uv_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the U and V parameters.

        Returns
        -------
        Vector3D
            Second derivative with respect to the U and V parameters.
        """
        return -np.sin(self.parameter.v) * self._torus.minor_radius.m * self.__cylinder_tangent

    @cached_property
    def vv_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            Second derivative with respect to the V parameter.
        """
        return (
            -np.cos(self.parameter.v) * self._torus.minor_radius.m * self.__cylinder_normal
            - np.sin(self.parameter.v) * self._torus.minor_radius.m * self._torus.dir_z
        )

    @cached_property
    def curvature(self) -> Tuple[Real, Vector3D, Real, Vector3D]:
        """
        Curvature of the torus.

        Returns
        -------
        Tuple[Real, Vector3D, Real, Vector3D]
            Minimum and maximum curvature value and direction, respectively.
        """
        min_cur = 1.0 / self._torus.minor_radius.m
        min_dir = UnitVector3D(self.v_derivative)
        start_point = self._torus.origin
        end_point = self.position
        max_cur = 1.0 / np.linalg.norm((end_point - start_point))
        max_dir = UnitVector3D(self.u_derivative)
        if min_cur > max_cur:
            min_cur, max_cur = max_cur, min_cur
            min_dir, max_dir = max_dir, min_dir
        return min_cur, min_dir, max_cur, max_dir

    @cached_property
    def min_curvature(self) -> Real:
        """
        Minimum curvature of the torus.

        Returns
        -------
        Real
            Minimum curvature of the torus.
        """
        return self.curvature[0]

    @cached_property
    def min_curvature_direction(self) -> UnitVector3D:
        """
        Minimum curvature direction.

        Returns
        -------
        UnitVector3D
            Minimum curvature direction.
        """
        return self.curvature[1]

    @cached_property
    def max_curvature(self) -> Real:
        """
        Maximum curvature of the torus.

        Returns
        -------
        Real
            Maximum curvature of the torus.
        """
        return self.curvature[2]

    @cached_property
    def max_curvature_direction(self) -> UnitVector3D:
        """
        Maximum curvature direction.

        Returns
        -------
        UnitVector3D
            Maximum curvature direction.
        """
        return self.curvature[3]



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a trapezoid."""

from beartype import beartype as check_input_types
from beartype.typing import Optional, Union
import numpy as np
from pint import Quantity
import pyvista as pv
from scipy.spatial.transform import Rotation as spatial_rotation

from ansys.geometry.core.math.constants import ZERO_POINT2D
from ansys.geometry.core.math.matrix import Matrix33
from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.segment import SketchSegment
from ansys.geometry.core.typing import Real


class Trapezoid(SketchFace):
    """
    Provides for modeling a 2D trapezoid.

    Parameters
    ----------
    width : Union[Quantity, Distance, Real]
        Width of the trapezoid.
    height : Union[Quantity, Distance, Real]
        Height of the trapezoid.
    slant_angle : Union[Quantity, Angle, Real]
        Angle for trapezoid generation.
    nonsymmetrical_slant_angle : Union[Quantity, Angle, Real], default: None
        Asymmetrical slant angles on each side of the trapezoid.
        The default is ``None``, in which case the trapezoid is symmetrical.
    center: Point2D, default: ZERO_POINT2D
        Center point of the trapezoid.
    angle : Union[Quantity, Angle, Real], default: 0
        Placement angle for orientation alignment.

    Notes
    -----
    If a nonsymmetrical slant angle is defined, the slant angle is
    applied to the left-most angle, and the nonsymmetrical slant angle
    is applied to the right-most angle.
    """

    @check_input_types
    def __init__(
        self,
        width: Union[Quantity, Distance, Real],
        height: Union[Quantity, Distance, Real],
        slant_angle: Union[Quantity, Angle, Real],
        nonsymmetrical_slant_angle: Optional[Union[Quantity, Angle, Real]] = None,
        center: Optional[Point2D] = ZERO_POINT2D,
        angle: Optional[Union[Quantity, Angle, Real]] = 0,
    ):
        """Initialize the trapezoid."""
        super().__init__()

        self._center = center
        self._width = width if isinstance(width, Distance) else Distance(width, center.unit)
        if self._width.value <= 0:
            raise ValueError("Width must be a real positive value.")
        width_magnitude = self._width.value.m_as(center.unit)

        self._height = height if isinstance(height, Distance) else Distance(height, center.unit)
        if self._height.value <= 0:
            raise ValueError("Height must be a real positive value.")
        height_magnitude = self._height.value.m_as(center.unit)

        if isinstance(angle, (int, float)):
            angle = Angle(angle, DEFAULT_UNITS.ANGLE)
        angle = angle if isinstance(angle, Angle) else Angle(angle, angle.units)

        if isinstance(slant_angle, (int, float)):
            slant_angle = Angle(slant_angle, DEFAULT_UNITS.ANGLE)
        slant_angle = (
            slant_angle if isinstance(slant_angle, Angle) else Angle(slant_angle, slant_angle.units)
        )

        if nonsymmetrical_slant_angle == None:
            nonsymmetrical_slant_angle = slant_angle
        else:
            if isinstance(nonsymmetrical_slant_angle, (int, float)):
                nonsymmetrical_slant_angle = Angle(nonsymmetrical_slant_angle, DEFAULT_UNITS.ANGLE)
            nonsymmetrical_slant_angle = (
                nonsymmetrical_slant_angle
                if isinstance(nonsymmetrical_slant_angle, Angle)
                else Angle(nonsymmetrical_slant_angle, nonsymmetrical_slant_angle.units)
            )

        rotation = Matrix33(
            spatial_rotation.from_euler(
                "xyz", [0, 0, angle.value.m_as(UNITS.radian)], degrees=False
            ).as_matrix()
        )

        half_h = height_magnitude / 2
        half_w = width_magnitude / 2
        rotated_point_1 = rotation @ [center.x.m - half_w, center.y.m - half_h, 0]
        rotated_point_2 = rotation @ [center.x.m + half_w, center.y.m - half_h, 0]
        rotated_point_3 = rotation @ [
            center.x.m - half_w + height_magnitude / np.tan(slant_angle.value.m_as(UNITS.radian)),
            center.y.m + half_h,
            0,
        ]
        rotated_point_4 = rotation @ [
            center.x.m
            + half_w
            - height_magnitude / np.tan(nonsymmetrical_slant_angle.value.m_as(UNITS.radian)),
            center.y.m + half_h,
            0,
        ]

        self._point1 = Point2D([rotated_point_1[0], rotated_point_1[1]], center.unit)
        self._point2 = Point2D([rotated_point_2[0], rotated_point_2[1]], center.unit)
        self._point3 = Point2D([rotated_point_3[0], rotated_point_3[1]], center.unit)
        self._point4 = Point2D([rotated_point_4[0], rotated_point_4[1]], center.unit)

        # TODO: add plane to SketchSegment when available
        self._segment1 = SketchSegment(self._point1, self._point2)
        self._segment2 = SketchSegment(self._point2, self._point3)
        self._segment3 = SketchSegment(self._point3, self._point4)
        self._segment4 = SketchSegment(self._point4, self._point1)

        self._edges.append(self._segment1)
        self._edges.append(self._segment2)
        self._edges.append(self._segment3)
        self._edges.append(self._segment4)

    @property
    def center(self) -> Point2D:
        """Center of the trapezoid."""
        return self._center

    @property
    def width(self) -> Quantity:
        """Width of the trapezoid."""
        return self._width.value

    @property
    def height(self) -> Quantity:
        """Height of the trapezoid."""
        return self._height.value

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """
        VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        import numpy as np

        return pv.Quadrilateral(
            np.array(
                [
                    [
                        self._point1.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._point1.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._point2.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._point2.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._point3.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._point3.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._point4.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._point4.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                ],
                dtype=np.float_,
            )
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing a bounding box."""

import sys

from beartype import beartype as check_input_types
from beartype.typing import List

from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.accuracy import Accuracy
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS
from ansys.geometry.core.typing import Real


class BoundingBox2D:
    """
    Maintains the X and Y dimensions.

    Parameters
    ----------
    x_min : Real
        Minimum value for the x-dimensional bounds.
    x_max : Real
        Maximum value for the x-dimensional bounds.
    y_min : Real
        Minimum value for the y-dimensional bounds.
    y_max : Real
        Maximum value for the y-dimensional bounds.
    """

    @check_input_types
    def __init__(
        self,
        x_min: Real = sys.float_info.max,
        x_max: Real = sys.float_info.min,
        y_min: Real = sys.float_info.max,
        y_max: Real = sys.float_info.min,
    ):
        """Initialize the ``BoundingBox2D`` class."""
        self._x_min = x_min
        self._x_max = x_max
        self._y_min = y_min
        self._y_max = y_max

    @property
    def x_min(self) -> Real:
        """
        Minimum value of X-dimensional bounds.

        Returns
        -------
        Real
            Minimum value of the X-dimensional bounds.
        """
        return self._x_min

    @property
    def x_max(self) -> Real:
        """
        Maximum value of the X-dimensional bounds.

        Returns
        -------
        Real
            Maximum value of the X-dimensional bounds.
        """
        return self._x_max

    @property
    def y_min(self) -> Real:
        """
        Minimum value of Y-dimensional bounds.

        Returns
        -------
        Real
            Minimum value of Y-dimensional bounds.
        """
        return self._y_min

    @property
    def y_max(self) -> Real:
        """
        Maximum value of Y-dimensional bounds.

        Returns
        -------
        Real
            Maximum value of Y-dimensional bounds.
        """
        return self._y_max

    @check_input_types
    def add_point(self, point: Point2D) -> None:
        """
        Extend the ranges of the bounding box to include a point.

        Notes
        -----
        This method is only applicable if the point components are outside
        the current bounds.

        Parameters
        ----------
        point : Point2D
            Point to include within the bounds.
        """
        self.add_point_components(
            point.x.m_as(DEFAULT_UNITS.LENGTH), point.y.m_as(DEFAULT_UNITS.LENGTH)
        )

    @check_input_types
    def add_point_components(self, x: Real, y: Real) -> None:
        """
        Extend the ranges of the bounding box to include the X and Y values.

        Notes
        -----
        This method is only applicable if the point components are outside
        the current bounds.

        Parameters
        ----------
        x : Real
            Point X component to include within the bounds.
        y : Real
            Point Y component to include within the bounds.
        """
        self._x_min = x if x < self._x_min else self._x_min
        self._x_max = x if x > self._x_max else self._x_max
        self._y_min = y if y < self._y_min else self._y_min
        self._y_max = y if y > self._y_max else self._y_max

    @check_input_types
    def add_points(self, points: List[Point2D]) -> None:
        """
        Extend the ranges of the bounding box to include given points.

        Parameters
        ----------
        points : List[Point2D]
            List of points to include within the bounds.
        """
        for point in points:
            self.add_point(point)

    @check_input_types
    def contains_point(self, point: Point2D) -> bool:
        """
        Evaluate whether a provided point lies within the X and Y ranges of the bounds.

        Parameters
        ----------
        point : Point2D
            Point to compare against the bounds.

        Returns
        -------
        bool
            ``True`` if the point is contained in the bounding box. Otherwise, ``False``.
        """
        return self.contains_point_components(
            point.x.m_as(DEFAULT_UNITS.LENGTH), point.y.m_as(DEFAULT_UNITS.LENGTH)
        )

    @check_input_types
    def contains_point_components(self, x: Real, y: Real) -> bool:
        """
        Check if point components are within current X and Y ranges of the bounds.

        Parameters
        ----------
        x : Real
            Point X component to compare against the bounds.
        y : Real
            Point Y component to compare against the bounds.

        Returns
        -------
        bool
            ``True`` if the components are contained in the bounding box. Otherwise, ``False``.
        """
        return (
            Accuracy.length_is_greater_than_or_equal(x, self._x_min)
            and Accuracy.length_is_greater_than_or_equal(y, self._y_min)
            and Accuracy.length_is_less_than_or_equal(x, self._x_max)
            and Accuracy.length_is_less_than_or_equal(y, self._y_max)
        )

    @check_input_types
    def __eq__(self, other: "BoundingBox2D") -> bool:
        """Equals operator for the ``BoundingBox2D`` class."""
        return (
            self.x_min == other.x_min
            and self.x_max == other.x_max
            and self.y_min == other.y_min
            and self.y_max == other.y_max
        )

    def __ne__(self, other: "BoundingBox2D") -> bool:
        """Not equals operator for the ``BoundingBox2D`` class."""
        return not self == other



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for evaluating decimal precision."""


import math

from ansys.geometry.core.typing import Real

LENGTH_ACCURACY: Real = 1e-8
"""Constant for decimal accuracy in length comparisons."""

ANGLE_ACCURACY: Real = 1e-6
"""Constant for decimal accuracy in angle comparisons."""

DOUBLE_ACCURACY: Real = 1e-13
"""Constant for double accuracy."""


class Accuracy:
    """Provides decimal precision evaluations for actions such as equivalency."""

    @staticmethod
    def length_accuracy() -> Real:
        """Return the ``LENGTH_ACCURACY`` constant."""
        return LENGTH_ACCURACY

    @staticmethod
    def angle_accuracy() -> Real:
        """Return the ``ANGLE_ACCURACY`` constant."""
        return ANGLE_ACCURACY

    @staticmethod
    def double_accuracy() -> Real:
        """Return the ``DOUBLE_ACCURACY`` constant."""
        return DOUBLE_ACCURACY

    @staticmethod
    def length_is_equal(comparison_length: Real, reference_length: Real) -> bool:
        """
        Check if the comparison length is equal to the reference length.

        Notes
        -----
        The check is done up to the constant value specified for ``LENGTH_ACCURACY``.

        Returns
        -------
        bool
            ``True`` if the comparison length is equal to the reference length
            within the length accuracy, ``False`` otherwise.
        """
        return Accuracy.is_within_tolerance(
            comparison_length, reference_length, LENGTH_ACCURACY, LENGTH_ACCURACY
        )

    @staticmethod
    def equal_doubles(a: Real, b: Real):
        """Compare two double values."""
        return Accuracy.is_within_tolerance(a, b, DOUBLE_ACCURACY, DOUBLE_ACCURACY)

    @staticmethod
    def compare_with_tolerance(
        a: Real, b: Real, relative_tolerance: Real, absolute_tolerance: Real
    ) -> Real:
        """Compare two doubles given the relative and absolute tolerances."""
        if Accuracy.is_within_tolerance(a, b, relative_tolerance, absolute_tolerance):
            return 0
        elif a < b:
            return -1
        else:
            return 1

    @staticmethod
    def length_is_greater_than_or_equal(comparison_length: Real, reference_length: Real) -> bool:
        """
        Check if the comparison length is greater than the reference length.

        Notes
        -----
        The check is done up to the constant value specified for ``LENGTH_ACCURACY``.

        Returns
        -------
        bool
            ``True`` if the comparison length is greater than the reference length within
            the length accuracy, ``False`` otherwise.
        """
        return bool(
            comparison_length > reference_length
            or Accuracy.length_is_equal(comparison_length, reference_length)
        )

    @staticmethod
    def length_is_less_than_or_equal(comparison_length: Real, reference_length: Real) -> bool:
        """
        Check if the comparison length is less than or equal to the reference length.

        Notes
        -----
        The check is done up to the constant value specified for ``LENGTH_ACCURACY``.

        Returns
        -------
        bool
            ``True`` if the comparison length is less than or equal to the reference length
            within the length accuracy, ``False`` otherwise.
        """
        return bool(
            comparison_length < reference_length
            or Accuracy.length_is_equal(comparison_length, reference_length)
        )

    @staticmethod
    def length_is_zero(length: Real) -> bool:
        """
        Check if the length is within the length accuracy of exact zero.

        Returns
        -------
        bool
            ``True`` if the length is within the length accuracy of exact zero,
            ``False`` otherwise.
        """
        return bool(length <= LENGTH_ACCURACY and length >= -LENGTH_ACCURACY)

    @staticmethod
    def length_is_negative(length: Real) -> bool:
        """
        Check if the length is below a negative length accuracy.

        Returns
        -------
        bool
            ``True`` if the length is below a negative length accuracy,
             ``False`` otherwise.
        """
        return bool(length < -LENGTH_ACCURACY)

    @staticmethod
    def length_is_positive(length: Real) -> bool:
        """
        Check if the length is above a positive length accuracy.

        Returns
        -------
        bool
            ``True`` if the length is above a positive length accuracy,
             ``False`` otherwise.
        """
        return bool(length > LENGTH_ACCURACY)

    @staticmethod
    def angle_is_zero(angle: Real) -> bool:
        """
        Check if the length is within the angle accuracy of exact zero.

        Returns
        -------
        bool
            ``True`` if the length is within the angle accuracy of exact zero,
             ``False`` otherwise.
        """
        return bool(abs(angle) < ANGLE_ACCURACY)

    @staticmethod
    def angle_is_negative(angle: Real) -> bool:
        """
        Check if the angle is below a negative angle accuracy.

        Returns
        -------
        bool
            ``True`` if the angle is below a negative angle accuracy,
             ``False`` otherwise.
        """
        return bool(angle <= -ANGLE_ACCURACY)

    @staticmethod
    def angle_is_positive(angle: Real) -> bool:
        """
        Check if the angle is above a positive angle accuracy.

        Returns
        -------
        bool
           ``True`` if the angle is above a positive angle accuracy,
            ``False`` otherwise.
        """
        return bool(angle >= ANGLE_ACCURACY)

    @staticmethod
    def is_within_tolerance(
        a: Real, b: Real, relative_tolerance: Real, absolute_tolerance: Real
    ) -> bool:
        """
        Check if two values (a and b) are inside a relative and absolute tolerance.

        Parameters
        ----------
        a : Real
            First value.
        b : Real
            Second value.
        relative_tolerance : Real
            Relative tolerance accepted.
        absolute_tolerance : Real
            Absolute tolerance accepted.

        Returns
        -------
        bool
            ``True`` if the values are inside the accepted tolerances,
            ``False`` otherwise.
        """
        # The code doesn't work for comparing infinity and non-infinite numbers!
        a_is_infinite = math.isinf(a)
        b_is_infinite = math.isinf(b)

        if a_is_infinite or b_is_infinite:
            if (a_is_infinite and a > 0) ^ (b_is_infinite and b > 0):
                return False
            if (a_is_infinite and a < 0) ^ (b_is_infinite and b < 0):
                return False

        difference = abs(a - b)
        return bool(
            difference < absolute_tolerance or difference <= relative_tolerance * 0.5 * abs(a + b)
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing a body."""
from abc import ABC, abstractmethod
from enum import Enum, unique
from functools import wraps

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.bodies_pb2 import (
    BooleanRequest,
    CopyRequest,
    GetCollisionRequest,
    MapRequest,
    MirrorRequest,
    RotateRequest,
    ScaleRequest,
    SetAssignedMaterialRequest,
    TranslateRequest,
)
from ansys.api.geometry.v0.bodies_pb2_grpc import BodiesStub
from ansys.api.geometry.v0.commands_pb2 import (
    AssignMidSurfaceOffsetTypeRequest,
    AssignMidSurfaceThicknessRequest,
    ImprintCurvesRequest,
    ProjectCurvesRequest,
)
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from beartype import beartype as check_input_types
from beartype.typing import TYPE_CHECKING, Iterable, List, Optional, Tuple, Union
from pint import Quantity

from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.conversions import (
    frame_to_grpc_frame,
    plane_to_grpc_plane,
    point3d_to_grpc_point,
    sketch_shapes_to_grpc_geometries,
    tess_to_pd,
    unit_vector_to_grpc_direction,
)
from ansys.geometry.core.designer.edge import CurveType, Edge
from ansys.geometry.core.designer.face import Face, SurfaceType
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.materials.material import Material
from ansys.geometry.core.math.constants import IDENTITY_MATRIX44
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.checks import check_type, ensure_design_is_active, min_backend_version
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.sketch.sketch import Sketch
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from pyvista import MultiBlock, PolyData

    from ansys.geometry.core.designer.component import Component


@unique
class MidSurfaceOffsetType(Enum):
    """Provides values for mid-surface offsets supported by the Geometry service."""

    MIDDLE = 0
    TOP = 1
    BOTTOM = 2
    VARIABLE = 3
    CUSTOM = 4


@unique
class CollisionType(Enum):
    """Provides values for collision types between bodies."""

    NONE = 0
    TOUCH = 1
    INTERSECT = 2
    CONTAINED = 3
    CONTAINEDTOUCH = 4


class IBody(ABC):
    """
    Defines the common methods for a body, providing the abstract body interface.

    Both the ``MasterBody`` class and ``Body`` class both inherit from the ``IBody``
    class. All child classes must implement all abstract methods.
    """

    @abstractmethod
    def id(self) -> str:
        """Get the ID of the body as a string."""
        return

    @abstractmethod
    def name(self) -> str:
        """Get the name of the body."""
        return

    @abstractmethod
    def faces(self) -> List[Face]:
        """
        Get a list of all faces within the body.

        Returns
        -------
        List[Face]
        """
        return

    @abstractmethod
    def edges(self) -> List[Edge]:
        """
        Get a list of all edges within the body.

        Returns
        -------
        List[Edge]
        """
        return

    @abstractmethod
    def is_alive(self) -> bool:
        """Check if the body is still alive and has not been deleted."""
        return

    @abstractmethod
    def is_surface(self) -> bool:
        """Check if the body is a planar body."""
        return

    @abstractmethod
    def surface_thickness(self) -> Union[Quantity, None]:
        """
        Get the surface thickness of a surface body.

        Notes
        -----
        This method is only for surface-type bodies that have been assigned a surface thickness.
        """
        return

    @abstractmethod
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:
        """
        Get the surface offset type of a surface body.

        Notes
        -----
        This method is only for surface-type bodies that have been assigned a surface offset.
        """
        return

    @abstractmethod
    def volume(self) -> Quantity:
        """
        Calculate the volume of the body.

        Notes
        -----
        When dealing with a planar surface, a value of ``0`` is returned as a volume.
        """
        return

    @abstractmethod
    def assign_material(self, material: Material) -> None:
        """
        Assign a material against the design in the active Geometry service instance.

        Parameters
        ----------
        material : Material
            Source material data.
        """
        return

    @abstractmethod
    def add_midsurface_thickness(self, thickness: Quantity) -> None:
        """
        Add a mid-surface thickness to a surface body.

        Parameters
        ----------
        thickness : ~pint.Quantity
            Thickness to assign.

        Notes
        -----
        Only surface bodies are eligible for mid-surface thickness assignment.
        """
        return

    @abstractmethod
    def add_midsurface_offset(self, offset: "MidSurfaceOffsetType") -> None:
        """
        Add a mid-surface offset to a surface body.

        Parameters
        ----------
        offset_type : MidSurfaceOffsetType
            Surface offset to assign.

        Notes
        -----
        Only surface bodies are eligible for mid-surface offset assignment.
        """
        return

    @abstractmethod
    def imprint_curves(self, faces: List[Face], sketch: Sketch) -> Tuple[List[Edge], List[Face]]:
        """
        Imprint all specified geometries onto specified faces of the body.

        Parameters
        ----------
        faces: List[Face]
            List of faces to imprint the curves of the sketch onto.
        sketch: Sketch
            All curves to imprint on the faces.

        Returns
        -------
        Tuple[List[Edge], List[Face]]
            All impacted edges and faces from the imprint operation.
        """
        return

    @abstractmethod
    def project_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: Optional[bool] = False,
    ) -> List[Face]:
        """
        Project all specified geometries onto the body.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the projection.
        sketch: Sketch
            All curves to project on the body.
        closest_face: bool
            Whether to target the closest face with the projection.
        only_one_curve: bool, default: False
            Whether to project only one curve of the entire sketch. When
            ``True``, only one curve is projected.

        Notes
        -----
        The ``only_one_curve`` parameter allows you to optimize the server call because
        projecting curves is an expensive operation. This reduces the workload on the
        server side.

        Returns
        -------
        List[Face]
            All faces from the project curves operation.
        """
        return

    @abstractmethod
    def imprint_projected_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: Optional[bool] = False,
    ) -> List[Face]:  # noqa: D102
        """
        Project and imprint specified geometries onto the body.

        This method combines the ``project_curves()`` and ``imprint_curves()`` method into
        one method. It has higher performance than calling them back-to-back when dealing
        with many curves. Because it is a specialized function, this method only returns
        the faces (and not the edges) from the imprint operation.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the projection.
        sketch: Sketch
            All curves to project on the body.
        closest_face: bool
            Whether to target the closest face with the projection.
        only_one_curve: bool, default: False
            Whether to project only one curve of the entire sketch. When
            ``True``, only one curve is projected.

        Notes
        -----
        The ``only_one_curve`` parameter allows you to optimize the server call because
        projecting curves is an expensive operation. This reduces the workload on the
        server side.

        Returns
        -------
        List[Face]
            All imprinted faces from the operation.
        """
        return

    @abstractmethod
    def translate(self, direction: UnitVector3D, distance: Union[Quantity, Distance, Real]) -> None:
        """
        Translate the geometry body in the specified direction by a given distance.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the translation.
        distance: Union[~pint.Quantity, Distance, Real]
            Distance (magnitude) of the translation.

        Returns
        -------
        None
        """
        return

    @abstractmethod
    def rotate(
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Union[Quantity, Angle, Real],
    ) -> None:
        """
        Rotate the geometry body around the specified axis by a given angle.

        Parameters
        ----------
        axis_origin: Point3D
            Origin of the rotational axis.
        axis_direction: UnitVector3D
            The axis of rotation.
        angle: Union[~pint.Quantity, Angle, Real]
            Angle (magnitude) of the rotation.

        Returns
        -------
        None
        """
        return

    @abstractmethod
    def scale(self, value: Real) -> None:
        """
        Scale the geometry body by the given value.

        Notes
        -----
        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.

        Parameters
        ----------
        value: Real
            Value to scale the body by.
        """
        return

    @abstractmethod
    def map(self, frame: Frame) -> None:
        """
        Map the geometry body to the new specified frame.

        Notes
        -----
        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.

        Parameters
        ----------
        frame: Frame
            Structure defining the orientation of the body.
        """
        return

    @abstractmethod
    def mirror(self, plane: Plane) -> None:
        """
        Mirror the geometry body across the specified plane.

        Notes
        -----
        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.

        Parameters
        ----------
        plane: Plane
            Represents the mirror.
        """
        return

    @abstractmethod
    def get_collision(self, body: "Body") -> CollisionType:
        """
        Get the collision state between bodies.

        Parameters
        ----------
        body: Body
            Object that the collision state is checked with.

        Returns
        -------
        CollisionType
            Enum that defines the collision state between bodies.
        """
        return

    @abstractmethod
    def copy(self, parent: "Component", name: str = None) -> "Body":
        """
        Create a copy of the body and place it under the specified parent component.

        Parameters
        ----------
        parent: Component
            Parent component to place the new body under within the design assembly.
        name: str
            Name to give the new body.

        Returns
        -------
        Body
            Copy of the body.
        """
        return

    @abstractmethod
    def tessellate(self, merge: Optional[bool] = False) -> Union["PolyData", "MultiBlock"]:
        """
        Tessellate the body and return the geometry as triangles.

        Parameters
        ----------
        merge : bool, default: False
            Whether to merge the body into a single mesh. When ``False`` (default), the
            number of triangles are preserved and only the topology is merged.
            When ``True``, the individual faces of the tessellation are merged.

        Returns
        -------
        ~pyvista.PolyData, ~pyvista.MultiBlock
            Merged :class:`pyvista.PolyData` if ``merge=True`` or a composite dataset.

        Examples
        --------
        Extrude a box centered at the origin to create a rectangular body and
        tessellate it:

        >>> from ansys.geometry.core.misc.units import UNITS as u
        >>> from ansys.geometry.core.sketch import Sketch
        >>> from ansys.geometry.core.math import Plane, Point2D, Point3D, UnitVector3D
        >>> from ansys.geometry.core import Modeler
        >>> modeler = Modeler()
        >>> origin = Point3D([0, 0, 0])
        >>> plane = Plane(origin, direction_x=[1, 0, 0], direction_y=[0, 0, 1])
        >>> sketch = Sketch(plane)
        >>> box = sketch.box(Point2D([2, 0]), 4, 4)
        >>> design = modeler.create_design("my-design")
        >>> my_comp = design.add_component("my-comp")
        >>> body = my_comp.extrude_sketch("my-sketch", sketch, 1 * u.m)
        >>> blocks = body.tessellate()
        >>> blocks
        >>> MultiBlock (0x7f94ec757460)
             N Blocks:	6
             X Bounds:	0.000, 4.000
             Y Bounds:	-1.000, 0.000
             Z Bounds:	-0.500, 4.500

        Merge the body:

        >>> mesh = body.tessellate(merge=True)
        >>> mesh
        PolyData (0x7f94ec75f3a0)
          N Cells:	12
          N Points:	24
          X Bounds:	0.000e+00, 4.000e+00
          Y Bounds:	-1.000e+00, 0.000e+00
          Z Bounds:	-5.000e-01, 4.500e+00
          N Arrays:	0
        """
        return

    @abstractmethod
    def plot(
        self,
        merge: bool = False,
        screenshot: Optional[str] = None,
        use_trame: Optional[bool] = None,
        **plotting_options: Optional[dict],
    ) -> None:
        """
        Plot the body.

        Parameters
        ----------
        merge : bool, default: False
            Whether to merge the body into a single mesh. When ``False`` (default),
            the number of triangles are preserved and only the topology is merged.
            When ``True``, the individual faces of the tessellation are merged.
        screenshot : str, default: None
            Path for saving a screenshot of the image that is being represented.
        use_trame : bool, default: None
            Whether to enable the use of `trame <https://kitware.github.io/trame/index.html>`_.
            The default is ``None``, in which case the ``USE_TRAME`` global setting
            is used.
        **plotting_options : dict, default: None
            Keyword arguments for plotting. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.

        Examples
        --------
        Extrude a box centered at the origin to create rectangular body and
        plot it:

        >>> from ansys.geometry.core.misc.units import UNITS as u
        >>> from ansys.geometry.core.sketch import Sketch
        >>> from ansys.geometry.core.math import Plane, Point2D, Point3D, UnitVector3D
        >>> from ansys.geometry.core import Modeler
        >>> modeler = Modeler()
        >>> origin = Point3D([0, 0, 0])
        >>> plane = Plane(origin, direction_x=[1, 0, 0], direction_y=[0, 0, 1])
        >>> sketch = Sketch(plane)
        >>> box = sketch.box(Point2D([2, 0]), 4, 4)
        >>> design = modeler.create_design("my-design")
        >>> mycomp = design.add_component("my-comp")
        >>> body = mycomp.extrude_sketch("my-sketch", sketch, 1 * u.m)
        >>> body.plot()

        Plot the body and color each face individually:

        >>> body.plot(multi_colors=True)
        """
        return

    def intersect(self, other: Union["Body", Iterable["Body"]]) -> None:
        """
        Intersect two (or more) bodies.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed.

        Parameters
        ----------
        other : Body
            Body to intersect with.

        Raises
        ------
        ValueError
            If the bodies do not intersect.
        """
        return

    @protect_grpc
    def subtract(self, other: Union["Body", Iterable["Body"]]) -> None:
        """
        Subtract two (or more) bodies.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed.

        Parameters
        ----------
        other : Body
            Body to subtract from the ``self`` parameter.

        Raises
        ------
        ValueError
            If the subtraction results in an empty (complete) subtraction.
        """
        return

    @protect_grpc
    def unite(self, other: Union["Body", Iterable["Body"]]) -> None:
        """
        Unite two (or more) bodies.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed.

        Parameters
        ----------
        other : Body
            Body to unite with the ``self`` parameter.
        """
        return


class MasterBody(IBody):
    """
    Represents solids and surfaces organized within the design assembly.

    Solids and surfaces synchronize to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    name : str
        User-defined label for the body.
    parent_component : Component
        Parent component to place the new component under within the design assembly.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    is_surface : bool, default: False
        Whether the master body is a surface or an 3D object (with volume). The default
        is ``False``, in which case the master body is a surface. When ``True``, the
        master body is a 3D object (with volume).
    """

    def __init__(
        self,
        id: str,
        name: str,
        grpc_client: GrpcClient,
        is_surface: bool = False,
    ):
        """Initialize the ``MasterBody`` class."""
        check_type(id, str)
        check_type(name, str)
        check_type(grpc_client, GrpcClient)
        check_type(is_surface, bool)

        self._id = id
        self._name = name
        self._grpc_client = grpc_client
        self._is_surface = is_surface
        self._surface_thickness = None
        self._surface_offset = None
        self._is_alive = True
        self._bodies_stub = BodiesStub(self._grpc_client.channel)
        self._commands_stub = CommandsStub(self._grpc_client.channel)
        self._tessellation = None

    def reset_tessellation_cache(func):
        """
        Decorate ``MasterBody`` methods that require a tessellation cache update.

        Parameters
        ----------
        func : method
            Method to call.

        Returns
        -------
        Any
            Output of the method, if any.
        """

        @wraps(func)
        def wrapper(self: "MasterBody", *args, **kwargs):
            self._tessellation = None
            return func(self, *args, **kwargs)

        return wrapper

    @property
    def _grpc_id(self) -> EntityIdentifier:  # noqa: D102
        """Entity identifier of this body on the server side."""
        return EntityIdentifier(id=self._id)

    @property
    def id(self) -> str:  # noqa: D102
        return self._id

    @property
    def name(self) -> str:  # noqa: D102
        return self._name

    @property
    def is_surface(self) -> bool:  # noqa: D102
        return self._is_surface

    @property
    def surface_thickness(self) -> Union[Quantity, None]:  # noqa: D102
        return self._surface_thickness if self.is_surface else None

    @property
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._surface_offset if self.is_surface else None

    @property
    @protect_grpc
    def faces(self) -> List[Face]:  # noqa: D102
        self._grpc_client.log.debug(f"Retrieving faces for body {self.id} from server.")
        grpc_faces = self._bodies_stub.GetFaces(self._grpc_id)
        return [
            Face(
                grpc_face.id,
                SurfaceType(grpc_face.surface_type),
                self,
                self._grpc_client,
                grpc_face.is_reversed,
            )
            for grpc_face in grpc_faces.faces
        ]

    @property
    @protect_grpc
    def edges(self) -> List[Edge]:  # noqa: D102
        self._grpc_client.log.debug(f"Retrieving edges for body {self.id} from server.")
        grpc_edges = self._bodies_stub.GetEdges(self._grpc_id)
        return [
            Edge(
                grpc_edge.id,
                CurveType(grpc_edge.curve_type),
                self,
                self._grpc_client,
                grpc_edge.is_reversed,
            )
            for grpc_edge in grpc_edges.edges
        ]

    @property
    def is_alive(self) -> bool:  # noqa: D102
        return self._is_alive

    @property
    @protect_grpc
    def volume(self) -> Quantity:  # noqa: D102
        if self.is_surface:
            self._grpc_client.log.debug("Dealing with planar surface. Returning 0 as the volume.")
            return Quantity(0, DEFAULT_UNITS.SERVER_VOLUME)
        else:
            self._grpc_client.log.debug(f"Retrieving volume for body {self.id} from server.")
            volume_response = self._bodies_stub.GetVolume(self._grpc_id)
            return Quantity(volume_response.volume, DEFAULT_UNITS.SERVER_VOLUME)

    @protect_grpc
    @check_input_types
    def assign_material(self, material: Material) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Assigning body {self.id} material {material.name}.")
        self._bodies_stub.SetAssignedMaterial(
            SetAssignedMaterialRequest(id=self._id, material=material.name)
        )

    @protect_grpc
    @check_input_types
    def add_midsurface_thickness(self, thickness: Quantity) -> None:  # noqa: D102
        if self.is_surface:
            self._commands_stub.AssignMidSurfaceThickness(
                AssignMidSurfaceThicknessRequest(
                    bodies_or_faces=[self.id], thickness=thickness.m_as(DEFAULT_UNITS.SERVER_LENGTH)
                )
            )
            self._surface_thickness = thickness
        else:
            self._grpc_client.log.warning(
                f"Body {self.name} cannot be assigned a mid-surface thickness because it is not a surface. Ignoring request."  # noqa : E501
            )

    @protect_grpc
    @check_input_types
    def add_midsurface_offset(self, offset: MidSurfaceOffsetType) -> None:  # noqa: D102
        if self.is_surface:
            self._commands_stub.AssignMidSurfaceOffsetType(
                AssignMidSurfaceOffsetTypeRequest(
                    bodies_or_faces=[self.id], offset_type=offset.value
                )
            )
            self._surface_offset = offset
        else:
            self._grpc_client.log.warning(
                f"Body {self.name} cannot be assigned a mid-surface offset because it is not a surface. Ignoring request."  # noqa : E501
            )

    @protect_grpc
    @check_input_types
    def imprint_curves(
        self, faces: List[Face], sketch: Sketch
    ) -> Tuple[List[Edge], List[Face]]:  # noqa: D102
        raise NotImplementedError(
            """
            imprint_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @protect_grpc
    @check_input_types
    def project_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: Optional[bool] = False,
    ) -> List[Face]:  # noqa: D102
        raise NotImplementedError(
            """
            project_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @check_input_types
    @protect_grpc
    def imprint_projected_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: Optional[bool] = False,
    ) -> List[Face]:  # noqa: D102
        raise NotImplementedError(
            """
            imprint_projected_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @protect_grpc
    @check_input_types
    @reset_tessellation_cache
    def translate(
        self, direction: UnitVector3D, distance: Union[Quantity, Distance, Real]
    ) -> None:  # noqa: D102
        distance = distance if isinstance(distance, Distance) else Distance(distance)

        translation_magnitude = distance.value.m_as(DEFAULT_UNITS.SERVER_LENGTH)

        self._grpc_client.log.debug(f"Translating body {self.id}.")

        self._bodies_stub.Translate(
            TranslateRequest(
                ids=[self.id],
                direction=unit_vector_to_grpc_direction(direction),
                distance=translation_magnitude,
            )
        )

    @protect_grpc
    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def rotate(
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Union[Quantity, Angle, Real],
    ) -> None:  # noqa: D102
        angle = angle if isinstance(angle, Angle) else Angle(angle)
        rotation_magnitude = angle.value.m_as(DEFAULT_UNITS.SERVER_ANGLE)
        self._grpc_client.log.debug(f"Rotating body {self.id}.")
        self._bodies_stub.Rotate(
            RotateRequest(
                id=self.id,
                axis_origin=point3d_to_grpc_point(axis_origin),
                axis_direction=unit_vector_to_grpc_direction(axis_direction),
                angle=rotation_magnitude,
            )
        )

    @protect_grpc
    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def scale(self, value: Real) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Scaling body {self.id}.")
        self._bodies_stub.Scale(ScaleRequest(id=self.id, scale=value))

    @protect_grpc
    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def map(self, frame: Frame) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Mapping body {self.id}.")
        self._bodies_stub.Map(MapRequest(id=self.id, frame=frame_to_grpc_frame(frame)))

    @protect_grpc
    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def mirror(self, plane: Plane) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Mirroring body {self.id}.")
        self._bodies_stub.Mirror(MirrorRequest(id=self.id, plane=plane_to_grpc_plane(plane)))

    @protect_grpc
    @min_backend_version(24, 2, 0)
    def get_collision(self, body: "Body") -> CollisionType:  # noqa: D102
        self._grpc_client.log.debug(f"Get collision between body {self.id} and body {body.id}.")
        response = self._bodies_stub.GetCollision(
            GetCollisionRequest(
                body_1_id=self.id,
                body_2_id=body.id,
            )
        )
        return CollisionType(response.collision)

    @protect_grpc
    def copy(self, parent: "Component", name: str = None) -> "Body":  # noqa: D102
        from ansys.geometry.core.designer.component import Component

        # Check input types
        check_type(parent, Component)
        check_type(name, (type(None), str))
        copy_name = self.name if name is None else name

        self._grpc_client.log.debug(f"Copying body {self.id}.")

        # Perform copy request to server
        response = self._bodies_stub.Copy(
            CopyRequest(
                id=self.id,
                parent=parent.id,
                name=copy_name,
            )
        )

        # Assign the new body to its specified parent (and return the new body)
        tb = MasterBody(
            response.master_id, copy_name, self._grpc_client, is_surface=self.is_surface
        )
        parent._master_component.part.bodies.append(tb)
        body_id = f"{parent.id}/{tb.id}" if parent.parent_component else tb.id
        return Body(body_id, response.name, parent, tb)

    @protect_grpc
    def tessellate(
        self, merge: Optional[bool] = False, transform: Matrix44 = IDENTITY_MATRIX44
    ) -> Union["PolyData", "MultiBlock"]:  # noqa: D102
        # lazy import here to improve initial module load time
        import pyvista as pv

        if not self.is_alive:
            return pv.PolyData() if merge else pv.MultiBlock()

        self._grpc_client.log.debug(f"Requesting tessellation for body {self.id}.")

        # cache tessellation
        if not self._tessellation:
            resp = self._bodies_stub.GetTessellation(self._grpc_id)
            self._tessellation = resp.face_tessellation.values()

        pdata = [tess_to_pd(tess).transform(transform) for tess in self._tessellation]
        comp = pv.MultiBlock(pdata)
        if merge:
            ugrid = comp.combine()
            return pv.PolyData(ugrid.points, ugrid.cells, n_faces=ugrid.n_cells)
        return comp

    def plot(
        self,
        merge: bool = False,
        screenshot: Optional[str] = None,
        use_trame: Optional[bool] = None,
        **plotting_options: Optional[dict],
    ) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement plot methods. Call this method on a body instead."
        )

    def intersect(self, other: Union["Body", Iterable["Body"]]) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def subtract(self, other: Union["Body", Iterable["Body"]]) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def unite(self, other: Union["Body", Iterable["Body"]]) -> None:
        # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def __repr__(self) -> str:
        """Represent the master body as a string."""
        lines = [f"ansys.geometry.core.designer.MasterBody {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(f"  Surface body         : {self.is_surface}")
        if self.is_surface:
            lines.append(f"  Surface thickness    : {self.surface_thickness}")
            lines.append(f"  Surface offset       : {self.surface_offset}")

        return "\n".join(lines)


class Body(IBody):
    """
    Represents solids and surfaces organized within the design assembly.

    Solids and surfaces synchronize to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    name : str
        User-defined label for the body.
    parent_component : Component
        Parent component to place the new component under within the design assembly.
    template : MasterBody
        Master body that this body is an occurrence of.
    """

    def __init__(self, id, name, parent_component: "Component", template: MasterBody) -> None:
        """Initialize the ``Body`` class."""
        self._id = id
        self._name = name
        self._parent_component = parent_component
        self._template = template

    def reset_tessellation_cache(func):
        """
        Decorate ``Body`` methods that require a tessellation cache update.

        Parameters
        ----------
        func : method
            Method to call.

        Returns
        -------
        Any
            Output of the method, if any.
        """

        @wraps(func)
        def wrapper(self: "Body", *args, **kwargs):
            self._template._tessellation = None
            return func(self, *args, **kwargs)

        return wrapper

    @property
    def id(self) -> str:  # noqa: D102
        return self._id

    @property
    def name(self) -> str:  # noqa: D102
        return self._template.name

    @property
    def parent_component(self) -> "Component":  # noqa: D102
        return self._parent_component

    @property
    @protect_grpc
    @ensure_design_is_active
    def faces(self) -> List[Face]:  # noqa: D102
        self._template._grpc_client.log.debug(f"Retrieving faces for body {self.id} from server.")
        grpc_faces = self._template._bodies_stub.GetFaces(EntityIdentifier(id=self.id))
        return [
            Face(
                grpc_face.id,
                SurfaceType(grpc_face.surface_type),
                self,
                self._template._grpc_client,
                grpc_face.is_reversed,
            )
            for grpc_face in grpc_faces.faces
        ]

    @property
    @protect_grpc
    @ensure_design_is_active
    def edges(self) -> List[Edge]:  # noqa: D102
        self._template._grpc_client.log.debug(f"Retrieving edges for body {self.id} from server.")
        grpc_edges = self._template._bodies_stub.GetEdges(EntityIdentifier(id=self.id))
        return [
            Edge(
                grpc_edge.id,
                CurveType(grpc_edge.curve_type),
                self,
                self._template._grpc_client,
                grpc_edge.is_reversed,
            )
            for grpc_edge in grpc_edges.edges
        ]

    @property
    def _is_alive(self) -> bool:  # noqa: D102
        return self._template.is_alive

    @_is_alive.setter
    def _is_alive(self, value: bool):
        self._template._is_alive = value

    @property
    def is_alive(self) -> bool:  # noqa: D102
        return self._is_alive

    @property
    def is_surface(self) -> bool:  # noqa: D102
        return self._template.is_surface

    @property
    def _surface_thickness(self) -> Union[Quantity, None]:  # noqa: D102
        return self._template.surface_thickness

    @_surface_thickness.setter
    def _surface_thickness(self, value):
        self._template._surface_thickness = value

    @property
    def surface_thickness(self) -> Union[Quantity, None]:  # noqa: D102
        return self._surface_thickness

    @property
    def _surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._template._surface_offset

    @_surface_offset.setter
    def _surface_offset(self, value: "MidSurfaceOffsetType"):  # noqa: D102
        self._template._surface_offset = value

    @property
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._surface_offset

    @property
    @ensure_design_is_active
    def volume(self) -> Quantity:  # noqa: D102
        return self._template.volume

    @ensure_design_is_active
    def assign_material(self, material: Material) -> None:  # noqa: D102
        self._template.assign_material(material)

    @ensure_design_is_active
    def add_midsurface_thickness(self, thickness: Quantity) -> None:  # noqa: D102
        self._template.add_midsurface_thickness(thickness)

    @ensure_design_is_active
    def add_midsurface_offset(self, offset: "MidSurfaceOffsetType") -> None:  # noqa: D102
        self._template.add_midsurface_offset(offset)

    @protect_grpc
    @ensure_design_is_active
    def imprint_curves(
        self, faces: List[Face], sketch: Sketch
    ) -> Tuple[List[Edge], List[Face]]:  # noqa: D102
        # Verify that each of the faces provided are part of this body
        body_faces = self.faces
        for provided_face in faces:
            is_found = False
            for body_face in body_faces:
                if provided_face.id == body_face.id:
                    is_found = True
                    break
            if not is_found:
                raise ValueError(f"Face with ID {provided_face.id} is not part of this body.")

        self._template._grpc_client.log.debug(
            f"Imprinting curves provided on {self.id} "
            + f"for faces {[face.id for face in faces]}."
        )

        imprint_response = self._template._commands_stub.ImprintCurves(
            ImprintCurvesRequest(
                body=self._id,
                curves=sketch_shapes_to_grpc_geometries(sketch._plane, sketch.edges, sketch.faces),
                faces=[face._id for face in faces],
            )
        )

        new_edges = [
            Edge(grpc_edge.id, CurveType(grpc_edge.curve_type), self, self._template._grpc_client)
            for grpc_edge in imprint_response.edges
        ]

        new_faces = [
            Face(
                grpc_face.id, SurfaceType(grpc_face.surface_type), self, self._template._grpc_client
            )
            for grpc_face in imprint_response.faces
        ]

        return (new_edges, new_faces)

    @protect_grpc
    @ensure_design_is_active
    def project_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: Optional[bool] = False,
    ) -> List[Face]:  # noqa: D102
        curves = sketch_shapes_to_grpc_geometries(
            sketch._plane, sketch.edges, sketch.faces, only_one_curve=only_one_curve
        )
        self._template._grpc_client.log.debug(f"Projecting provided curves on {self.id}.")

        project_response = self._template._commands_stub.ProjectCurves(
            ProjectCurvesRequest(
                body=self._id,
                curves=curves,
                direction=unit_vector_to_grpc_direction(direction),
                closest_face=closest_face,
            )
        )

        projected_faces = [
            Face(
                grpc_face.id, SurfaceType(grpc_face.surface_type), self, self._template._grpc_client
            )
            for grpc_face in project_response.faces
        ]

        return projected_faces

    @check_input_types
    @protect_grpc
    @ensure_design_is_active
    def imprint_projected_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: Optional[bool] = False,
    ) -> List[Face]:  # noqa: D102
        curves = sketch_shapes_to_grpc_geometries(
            sketch._plane, sketch.edges, sketch.faces, only_one_curve=only_one_curve
        )
        self._template._grpc_client.log.debug(f"Projecting provided curves on {self.id}.")

        response = self._template._commands_stub.ImprintProjectedCurves(
            ProjectCurvesRequest(
                body=self._id,
                curves=curves,
                direction=unit_vector_to_grpc_direction(direction),
                closest_face=closest_face,
            )
        )

        imprinted_faces = [
            Face(
                grpc_face.id, SurfaceType(grpc_face.surface_type), self, self._template._grpc_client
            )
            for grpc_face in response.faces
        ]

        return imprinted_faces

    @ensure_design_is_active
    def translate(
        self, direction: UnitVector3D, distance: Union[Quantity, Distance, Real]
    ) -> None:  # noqa: D102
        return self._template.translate(direction, distance)

    @ensure_design_is_active
    def rotate(
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Union[Quantity, Angle, Real],
    ) -> None:  # noqa: D102
        return self._template.rotate(axis_origin, axis_direction, angle)

    @ensure_design_is_active
    def scale(self, value: Real) -> None:  # noqa: D102
        return self._template.scale(value)

    @ensure_design_is_active
    def map(self, frame: Frame) -> None:  # noqa: D102
        return self._template.map(frame)

    @ensure_design_is_active
    def mirror(self, plane: Plane) -> None:  # noqa: D102
        return self._template.mirror(plane)

    @ensure_design_is_active
    def get_collision(self, body: "Body") -> CollisionType:  # noqa: D102
        return self._template.get_collision(body)

    @ensure_design_is_active
    def copy(self, parent: "Component", name: str = None) -> "Body":  # noqa: D102
        return self._template.copy(parent, name)

    @ensure_design_is_active
    def tessellate(
        self, merge: Optional[bool] = False
    ) -> Union["PolyData", "MultiBlock"]:  # noqa: D102
        return self._template.tessellate(merge, self.parent_component.get_world_transform())

    def plot(
        self,
        merge: bool = False,
        screenshot: Optional[str] = None,
        use_trame: Optional[bool] = None,
        show_options: Optional[dict] = None,
        **plotting_options: Optional[dict],
    ) -> None:  # noqa: D102
        # lazy import here to improve initial module load time
        from ansys.tools.visualization_interface.types.mesh_object_plot import MeshObjectPlot

        from ansys.geometry.core.plotting import GeometryPlotter

        meshobject = MeshObjectPlot(self, self.tessellate(merge=merge))
        pl = GeometryPlotter(use_trame=use_trame)
        pl.plot(meshobject, **plotting_options)
        pl.show(screenshot=screenshot, **plotting_options)

    def intersect(self, other: Union["Body", Iterable["Body"]]) -> None:  # noqa: D102
        self.__generic_boolean_op(other, "intersect", "bodies do not intersect")

    def subtract(self, other: Union["Body", Iterable["Body"]]) -> None:  # noqa: D102
        self.__generic_boolean_op(other, "subtract", "empty (complete) subtraction")

    def unite(self, other: Union["Body", Iterable["Body"]]) -> None:  # noqa: D102
        self.__generic_boolean_op(other, "unite", "union operation failed")

    @protect_grpc
    @reset_tessellation_cache
    @ensure_design_is_active
    @check_input_types
    def __generic_boolean_op(
        self, other: Union["Body", Iterable["Body"]], type_bool_op: str, err_bool_op: str
    ) -> None:
        grpc_other = other if isinstance(other, Iterable) else [other]
        try:
            response = self._template._bodies_stub.Boolean(
                BooleanRequest(
                    body1=self.id, tool_bodies=[b.id for b in grpc_other], method=type_bool_op
                )
            ).empty_result
        except Exception as err:
            # TODO: to be deleted - old versions did not have "tool_bodies" in the request
            # This is a temporary fix to support old versions of the server - should be deleted
            # once the server is no longer supported.
            if not isinstance(other, Iterable):
                response = self._template._bodies_stub.Boolean(
                    BooleanRequest(body1=self.id, body2=other.id, method=type_bool_op)
                ).empty_result
            else:
                all_response = []
                for body2 in other:
                    response = self._template._bodies_stub.Boolean(
                        BooleanRequest(body1=self.id, body2=body2.id, method=type_bool_op)
                    ).empty_result
                    all_response.append(response)

                if all_response.count(1) > 0:
                    response = 1

        if response == 1:
            raise ValueError(
                f"Boolean operation of type '{type_bool_op}' failed: {err_bool_op}.\n"
                f"Involving bodies:{self}, {grpc_other}"
            )

        for b in grpc_other:
            b.parent_component.delete_body(b)

    def __repr__(self) -> str:
        """Represent the ``Body`` as a string."""
        lines = [f"ansys.geometry.core.designer.Body {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(f"  Parent component     : {self._parent_component.name}")
        lines.append(f"  MasterBody           : {self._template.id}")
        lines.append(f"  Surface body         : {self.is_surface}")
        if self.is_surface:
            lines.append(f"  Surface thickness    : {self.surface_thickness}")
            lines.append(f"  Surface offset       : {self.surface_offset}")

        nl = "\n"
        return f"{nl}{nl.join(lines)}{nl}"



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a slot."""

from beartype import beartype as check_input_types
from beartype.typing import Optional, Union
import numpy as np
from pint import Quantity
import pyvista as pv
from scipy.spatial.transform import Rotation as spatial_rotation

from ansys.geometry.core.math.matrix import Matrix33
from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.arc import Arc
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.segment import SketchSegment
from ansys.geometry.core.typing import Real


class Slot(SketchFace):
    """
    Provides for modeling a 2D slot.

    Parameters
    ----------
    center: :class:`Point2D <ansys.geometry.core.math.point.Point2D>`
        Center point of the slot.
    width : Union[Quantity, Distance, Real]
        Width of the slot main body.
    height : Union[Quantity, Distance, Real]
        Height of the slot.
    angle : Union[Quantity, Angle, Real], default: 0
        Placement angle for orientation alignment.
    """

    @check_input_types
    def __init__(
        self,
        center: Point2D,
        width: Union[Quantity, Distance, Real],
        height: Union[Quantity, Distance, Real],
        angle: Optional[Union[Quantity, Angle, Real]] = 0,
    ):
        """Initialize the slot."""
        super().__init__()

        self._center = center
        self._width = width if isinstance(width, Distance) else Distance(width, center.unit)
        if self._width.value <= 0:
            raise ValueError("Width must be a real positive value.")
        width_magnitude = self._width.value.m_as(center.unit)

        self._height = height if isinstance(height, Distance) else Distance(height, center.unit)
        if self._height.value <= 0:
            raise ValueError("Height must be a real positive value.")
        height_magnitude = self._height.value.m_as(center.unit)

        if height_magnitude > width_magnitude:
            raise ValueError("Width must be greater than height.")

        if isinstance(angle, (int, float)):
            angle = Angle(angle, DEFAULT_UNITS.ANGLE)
        angle = angle if isinstance(angle, Angle) else Angle(angle, angle.units)

        rotation = Matrix33(
            spatial_rotation.from_euler(
                "xyz", [0, 0, angle.value.m_as(UNITS.radian)], degrees=False
            ).as_matrix()
        )

        half_h = height_magnitude / 2
        half_box_w = (width_magnitude - height_magnitude) / 2
        corner_1 = rotation @ [-half_box_w, half_h, 0]
        corner_2 = rotation @ [-half_box_w, -half_h, 0]
        corner_3 = rotation @ [half_box_w, -half_h, 0]
        corner_4 = rotation @ [half_box_w, half_h, 0]
        arc_1_center = rotation @ [-half_box_w, 0, 0]
        arc_2_center = rotation @ [half_box_w, 0, 0]

        self._slot_corner_1 = Point2D(
            [center.x.m + corner_1[0], center.y.m + corner_1[1]], center.unit
        )
        self._slot_corner_2 = Point2D(
            [center.x.m + corner_2[0], center.y.m + corner_2[1]], center.unit
        )
        self._slot_corner_3 = Point2D(
            [center.x.m + corner_3[0], center.y.m + corner_3[1]], center.unit
        )
        self._slot_corner_4 = Point2D(
            [center.x.m + corner_4[0], center.y.m + corner_4[1]], center.unit
        )
        self._arc_1_center = Point2D(
            [center.x.m + arc_1_center[0], center.y.m + arc_1_center[1]],
            center.unit,
        )
        self._arc_2_center = Point2D(
            [center.x.m + arc_2_center[0], center.y.m + arc_2_center[1]],
            center.unit,
        )

        # TODO: add plane to SketchSegment when available
        self._arc1 = Arc(self._slot_corner_1, self._slot_corner_2, self._arc_1_center)
        self._segment1 = SketchSegment(self._slot_corner_2, self._slot_corner_3)
        self._arc2 = Arc(self._slot_corner_3, self._slot_corner_4, self._arc_2_center)
        self._segment2 = SketchSegment(self._slot_corner_4, self._slot_corner_1)

        self._edges.append(self._arc1)
        self._edges.append(self._segment1)
        self._edges.append(self._arc2)
        self._edges.append(self._segment2)

    @property
    def center(self) -> Point2D:
        """Center of the slot."""
        return self._center

    @property
    def width(self) -> Quantity:
        """Width of the slot."""
        return self._width.value

    @property
    def height(self) -> Quantity:
        """Height of the slot."""
        return self._height.value

    @property
    def perimeter(self) -> Quantity:
        """Perimeter of the slot."""
        return np.pi * self._height.value + 2 * (self._width.value - self._height.value)

    @property
    def area(self) -> Quantity:
        """Area of the slot."""
        return (
            np.pi * (self._height.value / 2) ** 2
            + (self._width.value - self._height.value) * self._height.value
        )

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """
        VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        return pv.merge(
            [
                self._segment1.visualization_polydata,
                self._arc2.visualization_polydata,
                self._segment2.visualization_polydata,
                self._arc1.visualization_polydata,
            ]
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides plotting for various PyAnsys Geometry objects."""
from ansys.tools.visualization_interface import Color, EdgePlot, MeshObjectPlot, Plotter
from ansys.tools.visualization_interface.backends.pyvista import PyVistaBackend
from beartype.typing import Any, Dict, List, Optional, Union
import numpy as np
import pyvista as pv
from pyvista.plotting.tools import create_axes_marker

from ansys.geometry.core.designer.body import Body, MasterBody
from ansys.geometry.core.designer.component import Component
from ansys.geometry.core.designer.design import Design
from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.logger import LOG as logger
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.plotting.widgets import ShowDesignPoints
from ansys.geometry.core.sketch import Sketch


class GeometryPlotter(Plotter):
    """
    Plotter for PyAnsys Geometry objects.

    This class is an implementation of the PlotterInterface class.

    Parameters
    ----------
    use_trame : Union[bool, None], optional
        Whether to use trame visualizer or not, by default None.
    allow_picking : Union[bool, None], optional
        Whether to allow picking or not, by default False.
    show_plane : bool, optional
        Whether to show the plane in the scene, by default True.
    """

    def __init__(
        self,
        use_trame: Union[bool, None] = None,
        allow_picking: Union[bool, None] = False,
        show_plane: bool = True,
    ) -> None:
        """Initialize the GeometryPlotter class."""
        self._backend = PyVistaBackend(use_trame=use_trame, allow_picking=allow_picking)
        super().__init__(backend=self._backend)

        self._backend._allow_picking = allow_picking
        self._backend._use_trame = use_trame
        self._backend.add_widget(ShowDesignPoints(self))
        self._backend._pl._show_plane = show_plane

    def add_frame(self, frame: Frame, plotting_options: Optional[Dict] = None) -> None:
        """
        Plot a frame in the scene.

        Parameters
        ----------
        frame : Frame
            Frame to render in the scene.
        plotting_options : dict, default: None
            Dictionary containing parameters accepted by the
            :func:`pyvista.create_axes_marker` class for customizing
            the frame rendering in the scene.
        """
        # Use default plotting options if required
        if plotting_options is None:
            plotting_options = dict(labels_off=True, cone_radius=0.2)

        # Create the axes actor
        axes = create_axes_marker(**plotting_options)

        # Transpose the matrix for fixing rotation sense in VTK
        arr = np.vstack((frame.global_to_local_rotation, frame.origin)).T
        arr = np.vstack((arr, [0, 0, 0, 1]))

        # Apply matrix transformation to the actor
        axes.SetUserMatrix(pv.vtkmatrix_from_array(arr))

        # Render the actor in the scene
        self._backend.pv_interface.scene.add_actor(axes)

    def add_plane(
        self,
        plane: Plane,
        plane_options: Optional[Dict] = None,
        plotting_options: Optional[Dict] = None,
    ) -> None:
        """
        Plot a plane in the scene.

        Parameters
        ----------
        plane : Plane
            Plane to render in the scene.
        plane_options : dict, default: None
            Dictionary containing parameters accepted by the
            :func:`pyvista.Plane  <pyvista.Plane>` function for customizing the mesh
            representing the plane.
        plotting_options : dict, default: None
            Dictionary containing parameters accepted by the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method for
            customizing the mesh rendering of the plane.
        """
        # Impose default plane options if none provided
        if plane_options is None:
            plane_options = dict(i_size=10, j_size=10)

        plane_mesh = pv.Plane(
            center=plane.origin.tolist(), direction=plane.direction_z.tolist(), **plane_options
        )
        plane_meshobj = MeshObjectPlot(custom_object=plane, mesh=plane_mesh)
        # Impose default plotting options if none provided
        if plotting_options is None:
            plotting_options = dict(color="blue", opacity=0.1)

        self._backend.pv_interface.plot(plane_meshobj, **plotting_options)

    def add_sketch(
        self,
        sketch: Sketch,
        show_plane: bool = False,
        show_frame: bool = False,
        **plotting_options: Optional[Dict],
    ) -> None:
        """
        Plot a sketch in the scene.

        Parameters
        ----------
        sketch : Sketch
            Sketch to render in the scene.
        show_plane : bool, default: False
            Whether to render the sketch plane in the scene.
        show_frame : bool, default: False
            Whether to show the frame in the scene.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Show the sketch plane if required
        if show_plane:
            self.add_plane(sketch._plane)

        # Show the sketch plane if required
        if show_frame:
            self.add_frame(sketch._plane)

        if "clipping_plane" in plotting_options:
            logger.warning("Clipping is not available in Sketch objects.")
            plotting_options.pop("clipping_plane")

        self.add_sketch_polydata(
            sketch.sketch_polydata_faces(), sketch, opacity=0.7, **plotting_options
        )
        self.add_sketch_polydata(sketch.sketch_polydata_edges(), sketch, **plotting_options)

    def add_body_edges(self, body_plot: MeshObjectPlot, **plotting_options: Optional[dict]) -> None:
        """
        Add the outer edges of a body to the plot.

        This method has the side effect of adding the edges to the GeomObject that
        you pass through the parameters.

        Parameters
        ----------
        body_plot : MeshObjectPlot
            Body of which to add the edges.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        edge_plot_list = []
        for edge in body_plot.custom_object.edges:
            line = pv.Line(edge.start, edge.end)
            edge_actor = self._backend._pl.scene.add_mesh(
                line, line_width=10, color=Color.EDGE.value, **plotting_options
            )
            edge_actor.SetVisibility(False)
            edge_plot = EdgePlot(edge_actor, edge, body_plot)
            edge_plot_list.append(edge_plot)
        body_plot.edges = edge_plot_list

    def add_body(
        self, body: Body, merge: Optional[bool] = False, **plotting_options: Optional[Dict]
    ) -> None:
        """
        Add a body to the scene.

        Parameters
        ----------
        body : Body
            Body to add.
        merge : bool, default: False
            Whether to merge the body into a single mesh. When ``True``, the
            individual faces of the tessellation are merged. This
            preserves the number of triangles and only merges the topology.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments,
            see the :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Use the default PyAnsys Geometry add_mesh arguments
        self._backend.pv_interface.set_add_mesh_defaults(plotting_options)
        dataset = body.tessellate(merge=merge)
        body_plot = MeshObjectPlot(custom_object=body, mesh=dataset)
        self._backend.pv_interface.plot(body_plot, **plotting_options)
        self.add_body_edges(body_plot)

    def add_component(
        self,
        component: Component,
        merge_component: bool = False,
        merge_bodies: bool = False,
        **plotting_options,
    ) -> None:
        """
        Add a component to the scene.

        Parameters
        ----------
        component : Component
            Component to add.
        merge_component : bool, default: False
            Whether to merge the component into a single dataset. When
            ``True``, all the individual bodies are effectively combined
            into a single dataset without any hierarchy.
        merge_bodies : bool, default: False
            Whether to merge each body into a single dataset. When ``True``,
            all the faces of each individual body are effectively combined
            into a single dataset without separating faces.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Use the default PyAnsys Geometry add_mesh arguments
        self._backend.pv_interface.set_add_mesh_defaults(plotting_options)
        dataset = component.tessellate(merge_component=merge_component, merge_bodies=merge_bodies)
        component_polydata = MeshObjectPlot(component, dataset)
        self.plot(component_polydata, **plotting_options)

    def add_sketch_polydata(
        self, polydata_entries: List[pv.PolyData], sketch: Sketch = None, **plotting_options
    ) -> None:
        """
        Add sketches to the scene from PyVista polydata.

        Parameters
        ----------
        polydata_entries : List[pyvista.PolyData]
            Polydata to add.
        sketch : Sketch, default: None
            Sketch to add.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        # Use the default PyAnsys Geometry add_mesh arguments
        mb = pv.MultiBlock()
        for polydata in polydata_entries:
            mb.append(polydata)

        if sketch is None:
            self.plot(mb, color=Color.EDGE.value, **plotting_options)
        else:
            sk_polydata = MeshObjectPlot(custom_object=sketch, mesh=mb)
            self.plot(sk_polydata, color=Color.EDGE.value, **plotting_options)

    def add_design_point(self, design_point: DesignPoint, **plotting_options) -> None:
        """
        Add a DesignPoint object to the plotter.

        Parameters
        ----------
        design_point : DesignPoint
            DesignPoint to add.
        """
        design_point = MeshObjectPlot(custom_object=design_point, mesh=design_point._to_polydata())

        # get the actor for the DesignPoint
        self._backend.pv_interface.plot(design_point, **plotting_options)

    def plot_iter(
        self,
        plotting_list: List[Any],
        filter: str = None,
        **plotting_options,
    ) -> None:
        """
        Add a list of any type of object to the scene.

        These types of objects are supported: ``Body``, ``Component``, ``List[pv.PolyData]``,
        ``pv.MultiBlock``, and ``Sketch``.

        Parameters
        ----------
        plotting_list : List[Any]
            List of objects you want to plot.
        filter : str, default: None
            Regular expression with the desired name or names you want to include in the plotter.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        for object in plotting_list:
            _ = self.plot(object, filter, **plotting_options)

    # Override add function from plotter
    def plot(self, object: Any, filter: str = None, **plotting_options) -> None:
        """
        Add a custom mesh to the plotter.

        Parameters
        ----------
        object : Any
            Object to add.
        filter : str, default: None
            Regular expression with the desired name or names you want to include in the plotter.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, depend of the backend implementation
            you are using.
        """
        # Check for custom parameters in plotting_options
        if "merge_bodies" in plotting_options:
            merge_bodies = plotting_options["merge_bodies"]
            plotting_options.pop("merge_bodies", None)
        else:
            merge_bodies = None

        if "merge_components" in plotting_options:
            merge_components = plotting_options["merge_components"]
            plotting_options.pop("merge_components", None)
        else:
            merge_components = None
        # Add the custom object to the plotter
        if isinstance(object, DesignPoint):
            self.add_design_point(object, **plotting_options)
        elif isinstance(object, Sketch):
            self.add_sketch(object, **plotting_options)
        elif isinstance(object, Body) or isinstance(object, MasterBody):
            self.add_body(object, merge_bodies, **plotting_options)
        elif isinstance(object, Design) or isinstance(object, Component):
            self.add_component(object, merge_components, merge_bodies, **plotting_options)
        elif isinstance(object, List) and object != [] and isinstance(object[0], pv.PolyData):
            self.add_sketch_polydata(object, **plotting_options)
        elif isinstance(object, List):
            self.plot_iter(object, filter, **plotting_options)
        elif isinstance(object, MeshObjectPlot):
            self._backend.pv_interface.set_add_mesh_defaults(plotting_options)
            self._backend.pv_interface.plot(object, filter, **plotting_options)
        else:
            # any left type should be a PyVista object
            self._backend.pv_interface.plot(object, filter, **plotting_options)

    def show(
        self,
        plotting_object: Optional[Any] = None,
        screenshot: Optional[str] = None,
        **plotting_options,
    ) -> None:
        """
        Show the plotter.

        Parameters
        ----------
        plotting_object : Any, default: None
            Object you can add to the plotter.
        screenshot : str, default: None
            Path to save a screenshot of the plotter.
        **plotting_options : dict, default: None
            Keyword arguments for the plotter. Arguments depend of the backend implementation
            you are using.
        """
        if plotting_object is not None:
            self.plot(plotting_object, **plotting_options)
        self._backend.show(screenshot=screenshot, **plotting_options)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a line."""

from functools import cached_property
import math

from beartype import beartype as check_input_types
from beartype.typing import Union
import numpy as np

from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.accuracy import LENGTH_ACCURACY
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.curve_evaluation import CurveEvaluation
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
)
from ansys.geometry.core.typing import Real, RealSequence


class Line(Curve):
    """
    Provides 3D line representation.

    Parameters
    ----------
    origin : Union[~numpy.ndarray, RealSequence, Point3D]
        Origin of the line.
    direction : Union[~numpy.ndarray, RealSequence, UnitVector3D, Vector3D]
        Direction of the line.
    """

    @check_input_types
    def __init__(
        self,
        origin: Union[np.ndarray, RealSequence, Point3D],
        direction: Union[np.ndarray, RealSequence, UnitVector3D, Vector3D],
    ):
        """Initialize the ``Line`` class."""
        self._origin = Point3D(origin) if not isinstance(origin, Point3D) else origin
        self._direction = (
            UnitVector3D(direction) if not isinstance(direction, UnitVector3D) else direction
        )

    @property
    def origin(self) -> Point3D:
        """Origin of the line."""
        return self._origin

    @property
    def direction(self) -> UnitVector3D:
        """Direction of the line."""
        return self._direction

    def __eq__(self, other: object) -> bool:
        """Equals operator for the ``Line`` class."""
        if isinstance(other, Line):
            return self._origin == other.origin and self._direction == other.direction
        return False

    def evaluate(self, parameter: float) -> "LineEvaluation":
        """
        Evaluate the line at a given parameter.

        Parameters
        ----------
        parameter : Real
            Parameter to evaluate the line at.

        Returns
        -------
        LineEvaluation
            Resulting evaluation.
        """
        return LineEvaluation(self, parameter)

    def transformed_copy(self, matrix: Matrix44) -> "Line":
        """
        Create a transformed copy of the line based on a transformation matrix.

        Parameters
        ----------
        matrix : Matrix44
            4X4 transformation matrix to apply to the line.

        Returns
        -------
        Line
           New line that is the transformed copy of the original line.
        """
        old_origin_4d = np.array([[self.origin[0]], [self.origin[1]], [self.origin[2]], [1]])
        new_origin_4d = matrix * old_origin_4d
        new_point = Point3D([new_origin_4d[0], new_origin_4d[1], new_origin_4d[2]])
        new_axis = matrix * np.append(self._direction, 0)
        return Line(new_point, UnitVector3D(new_axis[0:3]))

    def project_point(self, point: Point3D) -> "LineEvaluation":
        """
        Project a point onto the line and evaluate the line.

        Parameters
        ----------
        point : Point3D
            Point to project onto the line.

        Returns
        -------
        LineEvaluation
            Resulting evaluation.
        """
        origin_to_point = point - self.origin
        t = origin_to_point.dot(self.direction)
        return LineEvaluation(self, t)

    def is_coincident_line(self, other: "Line") -> bool:
        """
        Determine if the line is coincident with another line.

        Parameters
        ----------
        other : Line
            Line to determine coincidence with.

        Returns
        -------
        bool
            ``True`` if the line is coincident with another line, ``False`` otherwise.
        """
        if self == other:
            return True
        if not self.direction.is_parallel_to(other.direction):
            return False

        between = Vector3D(self.origin - other.origin)

        return math.pow((self.direction % between).magnitude, 2) <= math.pow(
            LENGTH_ACCURACY, 2
        ) and math.pow((self.direction % between).magnitude, 2) <= math.pow(LENGTH_ACCURACY, 2)

    def is_opposite_line(self, other: "Line") -> bool:
        """
        Determine if the line is opposite another line.

        Parameters
        ----------
        other : Line
            Line to determine opposition with.

        Returns
        -------
        bool
            ``True`` if the line is opposite to another line.
        """
        if self.is_coincident_line(other):
            return self.direction.is_opposite(other.direction)
        return False

    def parameterization(self) -> Parameterization:
        """
        Get the parametrization of the line.

        The parameter of a line specifies the distance from the `origin` in the
        direction of `direction`.

        Returns
        -------
        Parameterization
            Information about how the line is parameterized.
        """
        return Parameterization(ParamForm.OPEN, ParamType.LINEAR, Interval(np.NINF, np.inf))

    def contains_param(self, param: Real) -> bool:  # noqa: D102
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:  # noqa: D102
        raise NotImplementedError("contains_point() is not implemented.")


class LineEvaluation(CurveEvaluation):
    """Provides for evaluating a line."""

    def __init__(self, line: Line, parameter: float = None) -> None:
        """Initialize the ``LineEvaluation`` class."""
        self._line = line
        self._parameter = parameter

    @property
    def line(self) -> Line:
        """Line being evaluated."""
        return self._line

    @property
    def parameter(self) -> float:
        """Parameter that the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """
        Position of the evaluation.

        Returns
        -------
        Point3D
            Point that lies on the line at this evaluation.
        """
        return self.line.origin + self.parameter * self.line.direction

    @cached_property
    def tangent(self) -> UnitVector3D:
        """
        Tangent of the evaluation, which is always equal to the direction of the line.

        Returns
        -------
        UnitVector3D
            Tangent unit vector to the line at this evaluation.
        """
        return self.line.direction

    @cached_property
    def first_derivative(self) -> Vector3D:
        """
        First derivative of the evaluation.

        The first derivative is always equal to the direction of the line.

        Returns
        -------
        Vector3D
            First derivative of the evaluation.
        """
        return self.line.direction

    @cached_property
    def second_derivative(self) -> Vector3D:
        """
        Second derivative of the evaluation.

        The second derivative is always equal to a zero vector ``Vector3D([0, 0, 0])``.

        Returns
        -------
        Vector3D
            Second derivative of the evaluation, which is always ``Vector3D([0, 0, 0])``.
        """
        return Vector3D([0, 0, 0])

    @cached_property
    def curvature(self) -> float:
        """
        Curvature of the line, which is always ``0``.

        Returns
        -------
        Real
            Curvature of the line, which is always ``0``.
        """
        return 0



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a cone."""

from functools import cached_property

from beartype import beartype as check_input_types
from beartype.typing import Tuple, Union
import numpy as np
from pint import Quantity

from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Z
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.measurements import Angle, Distance
from ansys.geometry.core.shapes.curves.line import Line
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
    ParamUV,
)
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.typing import Real, RealSequence


class Cone(Surface):
    """
    Provides 3D cone representation.

    Parameters
    ----------
    origin : Union[~numpy.ndarray, RealSequence, Point3D]
        Origin of the cone.
    radius : Union[Quantity, Distance, Real]
        Radius of the cone.
    half_angle : Union[Quantity, Angle, Real]
        Half angle of the apex, determining the upward angle.
    reference : Union[~numpy.ndarray, RealSequence, UnitVector3D, Vector3D]
        X-axis direction.
    axis : Union[~numpy.ndarray, RealSequence, UnitVector3D, Vector3D]
        Z-axis direction.
    """

    @check_input_types
    def __init__(
        self,
        origin: Union[np.ndarray, RealSequence, Point3D],
        radius: Union[Quantity, Distance, Real],
        half_angle: Union[Quantity, Angle, Real],
        reference: Union[np.ndarray, RealSequence, UnitVector3D, Vector3D] = UNITVECTOR3D_X,
        axis: Union[np.ndarray, RealSequence, UnitVector3D, Vector3D] = UNITVECTOR3D_Z,
    ):
        """Initialize the ``Cone`` class."""
        self._origin = Point3D(origin) if not isinstance(origin, Point3D) else origin
        self._reference = (
            UnitVector3D(reference) if not isinstance(reference, UnitVector3D) else reference
        )
        self._axis = UnitVector3D(axis) if not isinstance(axis, UnitVector3D) else axis
        if not self._reference.is_perpendicular_to(self._axis):
            raise ValueError("Cone reference (dir_x) and axis (dir_z) must be perpendicular.")

        self._radius = radius if isinstance(radius, Distance) else Distance(radius)
        if self._radius.value <= 0:
            raise ValueError("Radius must be a real positive value.")

        self._half_angle = half_angle if isinstance(half_angle, Angle) else Angle(half_angle)

    @property
    def origin(self) -> Point3D:
        """Origin of the cone."""
        return self._origin

    @property
    def radius(self) -> Quantity:
        """Radius of the cone."""
        return self._radius.value

    @property
    def half_angle(self) -> Quantity:
        """Half angle of the apex."""
        return self._half_angle.value

    @property
    def dir_x(self) -> UnitVector3D:
        """X-direction of the cone."""
        return self._reference

    @property
    def dir_y(self) -> UnitVector3D:
        """Y-direction of the cone."""
        return self.dir_z.cross(self.dir_x)

    @property
    def dir_z(self) -> UnitVector3D:
        """Z-direction of the cone."""
        return self._axis

    @property
    def height(self) -> Quantity:
        """Height of the cone."""
        return np.abs(self.radius / np.tan(self.half_angle))

    @property
    def surface_area(self) -> Quantity:
        """Surface area of the cone."""
        return np.pi * self.radius * (self.radius + np.sqrt(self.height**2 + self.radius**2))

    @property
    def volume(self) -> Quantity:
        """Volume of the cone."""
        return np.pi * self.radius**2 * self.height / 3

    def transformed_copy(self, matrix: Matrix44) -> "Cone":
        """
        Create a transformed copy of the cone based on a transformation matrix.

        Parameters
        ----------
        matrix : Matrix44
            4x4 transformation matrix to apply to the cone.

        Returns
        -------
        Cone
            New cone that is the transformed copy of the original cone.
        """
        new_point = self.origin.transform(matrix)
        new_reference = self._reference.transform(matrix)
        new_axis = self._axis.transform(matrix)
        return Cone(
            new_point,
            self.radius,
            self.half_angle,
            UnitVector3D(new_reference[0:3]),
            UnitVector3D(new_axis[0:3]),
        )

    def mirrored_copy(self) -> "Cone":
        """
        Create a mirrored copy of the cone along the y-axis.

        Returns
        -------
        Cone
            New cone that is a mirrored copy of the original cone.
        """
        return Cone(self.origin, self.radius, self.half_angle, -self._reference, -self._axis)

    @property
    def apex(self) -> Point3D:
        """Apex point of the cone."""
        return self.origin + self.apex_param * self.dir_z

    @property
    def apex_param(self) -> Real:
        """Apex parameter of the cone."""
        return -np.abs(self.radius.m) / np.tan(self.half_angle.m)

    @check_input_types
    def __eq__(self, other: "Cone") -> bool:
        """Equals operator for the ``Cone`` class."""
        return (
            self._origin == other._origin
            and self._radius == other._radius
            and self._half_angle == other._half_angle
            and self._reference == other._reference
            and self._axis == other._axis
        )

    def evaluate(self, parameter: ParamUV) -> "ConeEvaluation":
        """
        Evaluate the cone at given parameters.

        Parameters
        ----------
        parameter : ParamUV
            Parameters (u,v) to evaluate the cone at.

        Returns
        -------
        ConeEvaluation
            Resulting evaluation.
        """
        return ConeEvaluation(self, parameter)

    def project_point(self, point: Point3D) -> "ConeEvaluation":
        """
        Project a point onto the cone and evaluate the cone.

        Parameters
        ----------
        point : Point3D
            Point to project onto the cone.

        Returns
        -------
        ConeEvaluation
            Resulting evaluation.
        """
        u = np.arctan2(self.dir_y.dot(point - self.origin), self.dir_x.dot(point - self.origin))
        while u < 0:
            u += 2 * np.pi
        while u > 2 * np.pi:
            u -= 2 * np.pi
        axis = Line(self.origin, self.dir_z)
        line_eval = axis.project_point(point)
        v = line_eval.parameter

        cone_radius = self.radius.m + v * np.tan(self.half_angle.m)
        point_radius = np.linalg.norm(point - line_eval.position)
        dist_to_cone = (point_radius - cone_radius) * np.cos(self.half_angle.m)
        v += dist_to_cone * np.sin(self.half_angle.m)

        return ConeEvaluation(self, ParamUV(u, v))

    def parameterization(self) -> Tuple[Parameterization, Parameterization]:
        """
        Parameterize the cone surface as a tuple (U and V respectively).

        The U parameter specifies the clockwise angle around the axis (right-hand
        corkscrew law), with a zero parameter at ``dir_x`` and a period of 2*pi.

        The V parameter specifies the distance along the axis, with a zero parameter at
        the XY plane of the cone.

        Returns
        -------
        Tuple[Parameterization, Parameterization]
            Information about how a cone's u and v parameters are parameterized, respectively.
        """
        u = Parameterization(ParamForm.PERIODIC, ParamType.CIRCULAR, Interval(0, 2 * np.pi))

        start, end = (
            (self.apex_param, np.inf) if self.apex_param < 0 else (np.NINF, self.apex_param)
        )
        v = Parameterization(ParamForm.OPEN, ParamType.LINEAR, Interval(start, end))

        return (u, v)

    def contains_param(self, param_uv: ParamUV) -> bool:  # noqa: D102
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:  # noqa: D102
        raise NotImplementedError("contains_point() is not implemented.")


class ConeEvaluation(SurfaceEvaluation):
    """
    Evaluate the cone at given parameters.

    Parameters
    ----------
    cone: ~ansys.geometry.core.shapes.surfaces.cone.Cone
        Cone to evaluate.
    parameter: ParamUV
        Pparameters (u, v) to evaluate the cone at.
    """

    def __init__(self, cone: Cone, parameter: ParamUV) -> None:
        """Initialize the ``ConeEvaluation`` class."""
        self._cone = cone
        self._parameter = parameter

    @property
    def cone(self) -> Cone:
        """Cone being evaluated."""
        return self._cone

    @property
    def parameter(self) -> ParamUV:
        """Parameter that the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """
        Position of the evaluation.

        Returns
        -------
        Point3D
            Point that lies on the cone at this evaluation.
        """
        return (
            self.cone.origin
            + self.parameter.v * self.cone.dir_z
            + self.__radius_v * self.__cone_normal
        )

    @cached_property
    def normal(self) -> UnitVector3D:
        """
        Normal to the surface.

        Returns
        -------
        UnitVector3D
            Normal unit vector to the cone at this evaluation.
        """
        return UnitVector3D(
            self.__cone_normal * np.cos(self.cone.half_angle.m)
            - self.cone.dir_z * np.sin(self.cone.half_angle.m)
        )

    @cached_property
    def __radius_v(self) -> Real:
        """Private radius helper method."""
        return self.cone.radius.m + self.parameter.v * np.tan(self.cone.half_angle.m)

    @cached_property
    def __cone_normal(self) -> Vector3D:
        """Private normal helper method."""
        return (
            np.cos(self.parameter.u) * self.cone.dir_x + np.sin(self.parameter.u) * self.cone.dir_y
        )

    @cached_property
    def __cone_tangent(self) -> Vector3D:
        """Private tangent helper method."""
        return (
            -np.sin(self.parameter.u) * self.cone.dir_x + np.cos(self.parameter.u) * self.cone.dir_y
        )

    @cached_property
    def u_derivative(self) -> Vector3D:
        """
        First derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the U parameter.
        """
        return self.__radius_v * self.__cone_tangent

    @cached_property
    def v_derivative(self) -> Vector3D:
        """
        First derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the V parameter.
        """
        return self.cone.dir_z + np.tan(self.cone.half_angle.m) * self.__cone_normal

    @cached_property
    def uu_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            Second derivative with respect to the U parameter.
        """
        return -self.__radius_v * self.__cone_normal

    @cached_property
    def uv_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the U and V parameters.

        Returns
        -------
        Vector3D
            Second derivative with respect to U and V parameters.
        """
        return np.tan(self.cone.half_angle.m) * self.__cone_tangent

    @cached_property
    def vv_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            Second derivative with respect to the V parameter.
        """
        return Vector3D([0, 0, 0])

    @cached_property
    def min_curvature(self) -> Real:
        """
        Minimum curvature of the cone.

        Returns
        -------
        Real
            Minimum curvature of the cone.
        """
        return 0

    @cached_property
    def min_curvature_direction(self) -> UnitVector3D:
        """
        Minimum curvature direction.

        Returns
        -------
        UnitVector3D
            Minimum curvature direction.
        """
        return UnitVector3D(self.v_derivative)

    @cached_property
    def max_curvature(self) -> Real:
        """
        Maximum curvature of the cone.

        Returns
        -------
        Real
            Maximum curvature of the cone.
        """
        return 1.0 / self.__radius_v

    @cached_property
    def max_curvature_direction(self) -> UnitVector3D:
        """
        Maximum curvature direction.

        Returns
        -------
        UnitVector3D
            Maximum curvature direction.
        """
        return UnitVector3D(self.u_derivative)
